package com.fjhx.rabbitmq;

import com.alibaba.fastjson.JSONObject;
import com.fjhx.constant.RedisKeyConstant;
import com.fjhx.redis.RedisCache;
import org.springframework.amqp.AmqpException;
import org.springframework.amqp.core.Message;
import org.springframework.amqp.core.MessageBuilder;
import org.springframework.amqp.core.MessageProperties;
import org.springframework.amqp.rabbit.connection.CorrelationData;
import org.springframework.amqp.rabbit.core.RabbitTemplate;
import org.springframework.amqp.rabbit.core.RabbitTemplate.ConfirmCallback;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

import java.util.concurrent.TimeUnit;

/**
 * 消息发送接口实现类
 */
@Component
public class RabbitMQProducer implements ConfirmCallback {

    @Autowired
    private RabbitTemplate rabbitTemplate;

    @Autowired
    private RedisCache redisCache;

    @Value("${spring.rabbitmq.send.socket.exchange}")
    public String exchange;

    @Value("${spring.rabbitmq.send.socket.routingKey}")
    public String routingKey;

    @Value("${spring.rabbitmq.cache.time-out}")
    public Integer timeOut;

    /**
     * rabbitmq缓存key
     */
    private final String rabbitmqKey = RedisKeyConstant.RABBITMQ_REDIS_CACHE_KEY;

    /**
     * 消息发送机制
     *
     * @param data 内容
     */
    public void send(JSONObject data) {
        try {
            // 消息唯一id
            String messageId = data.getString("messageId");
            String msg = data.toJSONString();

            // 存入缓存
            redisCache.setCacheObject(rabbitmqKey + messageId, data, timeOut, TimeUnit.SECONDS);

            // 封装消息
            Message message = MessageBuilder.withBody(msg.getBytes())
                    .setContentType(MessageProperties.CONTENT_TYPE_JSON)
                    .setContentEncoding("UTF-8")
                    .setMessageId(messageId).build();

            // 发送消息
            this.rabbitTemplate.setMandatory(true);
            this.rabbitTemplate.setConfirmCallback(this);
            this.rabbitTemplate.convertAndSend(exchange, routingKey, message, new CorrelationData(messageId));

        } catch (AmqpException e) {
            e.printStackTrace();
        }
    }

    @Override
    public void confirm(CorrelationData correlationData, boolean ack, String cause) {
        // id 都是相同的哦 全局ID
        String messageId = correlationData.getId();
        // 消息发送失败，则继续发送
        if (!ack) {
            // 查询消息是否存在，当消息不存在得时候记录消息
            JSONObject jsonObject = redisCache.getCacheObject(rabbitmqKey + messageId);
            // 重试机制
            send(jsonObject);
        }
    }

}
