package com.fjhx.redis;

import com.fjhx.utils.StringUtils;
import lombok.extern.slf4j.Slf4j;
import org.redisson.api.RLock;
import org.redisson.api.RedissonClient;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.stereotype.Component;

import java.util.concurrent.TimeUnit;

/**
 * 分布式锁实现基于Redisson
 */
@Slf4j
@Component
public class RedissonLockClient {

    @Autowired
    private RedissonClient redissonClient;

    @Autowired
    private RedisTemplate redisTemplate;

    /**
     * 获取锁
     *
     * @param lockKey 锁名称
     * @return
     */
    public RLock getLock(String lockKey) {
        return redissonClient.getLock(lockKey);
    }

    /**
     * 尝试加锁
     *
     * @param lockKey   锁名称
     * @param leaseTime 如果锁获取成功，这把锁的过期时间，单位分钟
     * @return
     */
    public boolean tryLock(String lockKey, long leaseTime) {
        return tryLock(lockKey, 0, leaseTime);
    }

    /**
     * 尝试加锁
     *
     * @param lockKey   锁名称
     * @param waitTime  获取锁的最长等待时间
     * @param leaseTime 如果锁获取成功，这把锁的过期时间，单位分钟
     * @return
     */
    public boolean tryLock(String lockKey, long waitTime, long leaseTime) {
        RLock rLock = getLock(lockKey);
        boolean getLock = false;
        try {
            getLock = rLock.tryLock(waitTime, leaseTime, TimeUnit.MINUTES);
            if (getLock) {
                log.debug("获取锁成功,lockKey={}", lockKey);
            } else {
                log.debug("获取锁失败,lockKey={}", lockKey);
            }
        } catch (Exception e) {
            log.error("获取式锁异常，lockKey=" + lockKey, e);
            getLock = false;
        }
        return getLock;
    }

    /**
     * 尝试加锁
     *
     * @param lockKey
     * @return
     */
    public boolean tryLock(String lockKey) {
        RLock rLock = getLock(lockKey);
        boolean getLock = false;
        try {
            getLock = rLock.tryLock();
            if (getLock) {
                log.debug("获取锁成功,lockKey={}", lockKey);
            } else {
                log.debug("获取锁失败,lockKey={}", lockKey);
            }
        } catch (Exception e) {
            log.error("获取式锁异常，lockKey=" + lockKey, e);
            getLock = false;
        }
        return getLock;
    }

    /**
     * 尝试获取锁（实现公平锁定，从而保证线程的获取顺序）<br/>
     * 使用公平锁，如果获取了锁，则为true，否则为false
     *
     * @param lockKey 锁名称
     * @return
     */
    public boolean getFairTryLock(String lockKey) {
        RLock fairLock = redissonClient.getFairLock(lockKey);
        try {
            boolean tryLock = fairLock.tryLock();
            if (!tryLock) {
                log.debug("锁已经存在了,lockKey={}", lockKey);
            }
            return tryLock;
        } catch (Exception e) {
            e.printStackTrace();
        }
        return false;
    }

    /**
     * 公平锁（实现公平锁定，从而保证线程的获取顺序）
     *
     * @param lockKey   锁名称
     * @param leaseTime 锁的过期时间
     * @param unit      时间单位
     * @return
     */
    public boolean getFairLock(String lockKey, int leaseTime, TimeUnit unit) {
        RLock fairLock = redissonClient.getFairLock(lockKey);
        try {
            boolean existKey = existKey(lockKey);
            // 已经存在了，就直接返回
            if (existKey) {
                log.debug("锁已经存在了,lockKey={}", lockKey);
                return false;
            }
            return fairLock.tryLock(3, leaseTime, unit);
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        return false;
    }

    /**
     * 公平锁（实现公平锁定，从而保证线程的获取顺序）
     *
     * @param lockKey 锁名称
     * @return
     */
    public RLock getFairLock(String lockKey) {
        RLock fairLock = redissonClient.getFairLock(lockKey);
        return fairLock;
    }

    /**
     * 缓存key是否存在
     *
     * @param key key名称
     * @return
     */
    public boolean existKey(String key) {
        return redisTemplate.hasKey(key);
    }

    /**
     * 加锁
     *
     * @param lockKey 锁名称
     * @return
     */
    public RLock lock(String lockKey) {
        RLock lock = getLock(lockKey);
        lock.lock();
        return lock;
    }

    /**
     * 加锁
     *
     * @param lockKey   锁名称
     * @param leaseTime 锁的过期时间
     * @return
     */
    public RLock lock(String lockKey, long leaseTime) {
        RLock lock = getLock(lockKey);
        lock.lock(leaseTime, TimeUnit.SECONDS);
        return lock;
    }

    /**
     * 解锁
     *
     * @param lockKey 锁名称
     */
    public void unlock(String lockKey) {
        try {
            if (StringUtils.isBlank(lockKey)) {
                return;
            }
            redissonClient.getLock(lockKey).unlock();
        } catch (Exception e) {

        }
    }

    /**
     * 公平锁，解锁
     *
     * @param lockKey
     */
    public void fairUnlock(String lockKey) {
        if (StringUtils.isBlank(lockKey)) {
            return;
        }
        redissonClient.getFairLock(lockKey).unlock();
    }

}
