package com.fjhx.rfid.r2000.udp.service.impl;

import com.alibaba.fastjson.JSONObject;
import com.fjhx.constant.RFIDConstant;
import com.fjhx.constant.RedisKeyConstant;
import com.fjhx.entity.RFIDClient;
import com.fjhx.redis.RedisCache;
import com.fjhx.rfid.r2000.udp.callback.UdpCallBackSaveLog;
import com.fjhx.rfid.r2000.udp.service.UdpService;
import com.fjhx.rfid.r2000.udp.utils.Utility;
import com.fjhx.thread.ThreadPoolManager;
import com.fjhx.utils.AjaxResult;
import com.fjhx.web.exception.BaseException;
import io.netty.bootstrap.Bootstrap;
import io.netty.buffer.ByteBuf;
import io.netty.buffer.PooledByteBufAllocator;
import io.netty.buffer.Unpooled;
import io.netty.channel.Channel;
import io.netty.channel.ChannelOption;
import io.netty.channel.EventLoopGroup;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.DatagramPacket;
import io.netty.channel.socket.nio.NioDatagramChannel;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.collections4.MapUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.net.InetSocketAddress;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CountDownLatch;
import java.util.stream.Collectors;

/**
 * @Description:
 * @ClassName: UdpServiceImpl
 * @Author: linqt
 * @Date: 2022/3/18 11:22
 * @Version: 1.0
 */
@Service
public class UdpServiceImpl implements UdpService {

    @Autowired
    private RedisCache redisCache;

    /**
     * 初始化实例集合
     */
    private static Map<String, Object> serverMap = new HashMap<>();

    /**
     * 启动单台
     *
     * @param client
     * @return
     */
    public AjaxResult startOne(RFIDClient client) {
        // redis udp数据缓存key
        String dataRedisKey = MessageFormat.format(RedisKeyConstant.RFID_UDP_DATA_REDIS_CACHE_PREFIX, client.getIp());

        try {
            EventLoopGroup group = new NioEventLoopGroup();
            Bootstrap b = new Bootstrap();
            b.group(group)
                    .channel(NioDatagramChannel.class)
                    .option(ChannelOption.SO_BROADCAST, true)
                    .option(ChannelOption.ALLOCATOR, PooledByteBufAllocator.DEFAULT)
                    .handler(new UdpCallBackSaveLog(dataRedisKey, client.getUserName()));
            Channel ch = b.bind(RFIDConstant.RFID_UDP_DEFAULT_BINDING_PORT).sync().channel();
            // 读取
            byte[] sendData = Utility.convert2HexArray(RFIDConstant.RFID_UDP_START_READ_INSTRUCTION);
            ByteBuf byteBuf = Unpooled.copiedBuffer(sendData);
            ch.writeAndFlush(new DatagramPacket(byteBuf, new InetSocketAddress(client.getIp(), client.getPort()))).sync();

            // 实例静态化
            serverMap.put(client.getIp(), ch);

        } catch (Exception e) {
            e.printStackTrace();
            throw new BaseException(RFIDConstant.SERVER_START_ERR);
        }
        return AjaxResult.success(RFIDConstant.SERVER_START_SUCCESS);
    }

    /**
     * 关闭单台
     *
     * @param client
     * @return
     */
    public AjaxResult closeOne(RFIDClient client) {
        Map<String, JSONObject> data = new HashMap<>();
        List<JSONObject> list = new ArrayList<>();

        // redis udp数据缓存key
        String dataRedisKey = MessageFormat.format(RedisKeyConstant.RFID_UDP_DATA_REDIS_CACHE_PREFIX, client.getIp());

        try {
            // 从静态获取实例
            Channel ch = (Channel) serverMap.get(client.getIp());
            // 停止读取
            byte[] sendData = Utility.convert2HexArray(RFIDConstant.RFID_UDP_STOP_READ_INSTRUCTION);
            ByteBuf byteBuf = Unpooled.copiedBuffer(sendData);
            ch.writeAndFlush(new DatagramPacket(byteBuf, new InetSocketAddress(client.getIp(), client.getPort()))).sync();
            ch.close();

            // 移除服务
            serverMap.remove(client.getIp());

            // 获取数据缓存
            data = redisCache.getCacheObject(dataRedisKey);
//            System.err.println("当前缓存中取出的数据量:" + data.size());

            // 删除数据缓存
            redisCache.deleteObject(dataRedisKey);

            if (MapUtils.isNotEmpty(data)) {
                list = data.values().stream().collect(Collectors.toList());
            }

        } catch (Exception e) {
            e.printStackTrace();
            throw new BaseException(RFIDConstant.SERVER_CLOSE_ERR);
        }
        return AjaxResult.success(RFIDConstant.SERVER_CLOSE_SUCCESS, list);
    }

    /**
     * 启动多台
     *
     * @param client
     * @return
     */
    @Override
    public AjaxResult openMore(RFIDClient client) {
        ConcurrentHashMap<String, Object> sMap = new ConcurrentHashMap<>();
        ConcurrentHashMap<String, JSONObject> dataMap = new ConcurrentHashMap<>();
        List<JSONObject> list = new ArrayList<>();

        Integer count = 0;
        try {
            if (client == null || CollectionUtils.isEmpty(client.getIpList())) {
                throw new BaseException("读写器IP地址不能为空！");
            }
            CountDownLatch countDownLatch = new CountDownLatch(client.getIpList().size());
            // 启动扫描线程
            for (String ip : client.getIpList()) {
                count++;
                Integer finalCount = count;
                CountDownLatch errCountDownLatch = countDownLatch;
                ThreadPoolManager.execute(() -> {
                    try {
                        // redis udp数据缓存key
                        String dataRedisKey = MessageFormat.format(RedisKeyConstant.RFID_UDP_DATA_REDIS_CACHE_PREFIX, ip + "-" + finalCount);
                        EventLoopGroup group = new NioEventLoopGroup();
                        Bootstrap b = new Bootstrap();
                        b.group(group)
                                .channel(NioDatagramChannel.class)
                                .option(ChannelOption.SO_BROADCAST, true)
                                .option(ChannelOption.ALLOCATOR, PooledByteBufAllocator.DEFAULT)
                                .handler(new UdpCallBackSaveLog(dataRedisKey, null));
                        Channel ch = b.bind(RFIDConstant.RFID_UDP_DEFAULT_BINDING_PORT).sync().channel();
                        // 读取
                        byte[] sendData = Utility.convert2HexArray(RFIDConstant.RFID_UDP_START_READ_INSTRUCTION);
                        ByteBuf byteBuf = Unpooled.copiedBuffer(sendData);
                        ch.writeAndFlush(new DatagramPacket(byteBuf, new InetSocketAddress(ip, client.getPort()))).sync();

                        // 实例静态化
                        sMap.put(ip + "-" + finalCount, ch);
                    } catch (Exception e) {
                        e.printStackTrace();
                        if (errCountDownLatch != null) {
                            errCountDownLatch.countDown();
                        }
                        throw new BaseException(RFIDConstant.READ_DATA_ERR);
                    }
                });
            }

            // 读取时间
            Thread.sleep(client.getReadTime());

            count = 0;
            // 关闭线程
            for (String ip : client.getIpList()) {
                count++;
                Integer finalCount = count;
                CountDownLatch finalCountDownLatch = countDownLatch;
                try {
                    // redis udp数据缓存key
                    String dataRedisKey = MessageFormat.format(RedisKeyConstant.RFID_UDP_DATA_REDIS_CACHE_PREFIX, ip + "-" + finalCount);

                    // 获取实例
                    Channel ch = (Channel) sMap.get(ip + "-" + finalCount);
                    // 停止读取
                    byte[] sendData = Utility.convert2HexArray(RFIDConstant.RFID_UDP_STOP_READ_INSTRUCTION);
                    ByteBuf byteBuf = Unpooled.copiedBuffer(sendData);
                    ch.writeAndFlush(new DatagramPacket(byteBuf, new InetSocketAddress(ip, client.getPort()))).sync();
                    ch.close();

                    // 移除服务
                    sMap.remove(ip + "-" + finalCount);

                    // 获取数据缓存
                    ConcurrentHashMap<String, JSONObject> data = redisCache.getCacheObject(dataRedisKey);
                    if (MapUtils.isEmpty(data)) {
//                        System.err.println(ip + "-" + finalCount + "当前缓存中取出的数据为空！");
                        continue;
                    }
//                    System.err.println(ip + "-" + finalCount + "当前缓存中取出的数据量：" + data.size());
                    dataMap.putAll(data);

                    // 删除数据缓存
                    redisCache.deleteObject(dataRedisKey);
                } catch (Exception e) {
                    e.printStackTrace();
                } finally {
                    if (finalCountDownLatch != null) {
                        finalCountDownLatch.countDown();
                    }
                }
            }
//            System.err.println("最终总结果数量--:" + dataMap.size());

            if (MapUtils.isNotEmpty(dataMap)) {
                list = dataMap.values().stream().collect(Collectors.toList());
            }

            if (countDownLatch != null) {
                countDownLatch.await();
            }
        } catch (Exception e) {
            e.printStackTrace();
        }

        return AjaxResult.success(RFIDConstant.READ_DATA_SUCCESS, list);
    }
}
